from __future__ import print_function,division
import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import simpledist.distributions as dists
from plotutils import setfig
import scipy.stats as stats
from scipy.interpolate import UnivariateSpline as interpolate
from scipy.special import erf
from numpy import pi

from .kappa_inference import cosi_posterior,sample_posterior

from astropy import constants as const
RSUN = const.R_sun.cgs.value
MSUN = const.M_sun.cgs.value
DAY = 86400

class Cosi_Distribution(dists.Distribution):
    """A Distribution for cos(I*) given Rstar, Prot, Vsini measurements

    Measurements may be passed as either (val,err) tuples/lists or 
    as `distribution` (or `stats.continuous`) objects

    if vsini_dist is passed as a tuple/list, and vsini_corrected is False,
    then by default the value of vsini will be corrected by dividing by 
    1 - (alpha/2), to correct for differential rotation, as in 
    Hirano et al. (2013/14)
    """
    def __init__(self,R_dist,Prot_dist,vsini_dist,nsamples=1e3,
                 vgrid=None,npts=100,vgrid_pts=1000,vsini_corrected=False,
                 veq_simple=False,
                 N_veq_samples=1e3,alpha=0.23,l0=20,sigl=20,
                 veq_bandwidth=0.1):

        if type(R_dist) == type(''):
            R_dist = dists.Distribution_FromH5(R_dist,'radius')
        elif type(R_dist) in [type([]),type((1,))]:
            if len(R_dist)==2:
                R,dR = R_dist
                R=float(R); dR=float(dR)
                R_dist = dists.Gaussian_Distribution(R,dR,name='radius')
            elif len(R_dist)==3:
                R_dist = dists.fit_doublegauss(float(R_dist[0]),float(R_dist[1]),float(R_dist[2]),name='radius')
        if type(Prot_dist) in [type([]),type((1,))]:
            P,dP = Prot_dist
            P=float(P); dP=float(dP)
            Prot_dist = dists.Gaussian_Distribution(P,dP,name='Prot')
        if type(vsini_dist) in [type([]),type((1,))]:
            vsini,dvsini = vsini_dist
            vsini=float(vsini); dvsini=float(dvsini)
            if not vsini_corrected:
                vsini = float(vsini)/(1 - (alpha/2))
            vsini_dist = dists.Gaussian_Distribution(vsini,dvsini,name='vsini')
        if vsini_dist.minval < 0:
            vsini_dist.minval = 0


        if veq_simple:
            veq_dist = Veq_Distribution_Simple(R,dR,P,dP)
        else:
            veq_dist = Veq_Distribution(R_dist,Prot_dist,N=N_veq_samples,
                                        alpha=alpha,l0=l0,sigl=sigl,
                                        bandwidth=veq_bandwidth)
        self.vsini_dist = vsini_dist
        self.veq_dist = veq_dist


        cs,Ls = cosi_posterior(vsini_dist, veq_dist, vgrid=vgrid,
                             npts=npts, vgrid_pts=vgrid_pts)
        self.samples = sample_posterior(cs,Ls,nsamples=nsamples)
        pdf = interpolate(cs,Ls,s=0)
        
        dists.Distribution.__init__(self,pdf,name='cos(I)',
                                    minval=0,maxval=1)

    def save_hdf(self,filename,path='',**kwargs):
        dists.Distribution.save_hdf(self,filename,path,**kwargs)
        self.vsini_dist.save_hdf(filename,'{}/vsini'.format(path))
        self.veq_dist.save_hdf(filename,'{}/veq'.format(path))

    def summary_plot(self,fig=None,title=''):
        setfig(fig)
        
        ax1 = plt.subplot(221)
        ax2 = plt.subplot(222)
        ax3 = plt.subplot(223)

        plt.sca(ax1)
        self.plot(fig=0)
        plt.yticks([])
        plt.xlabel('Cos(I)',labelpad=2)
        plt.sca(ax2)
        self.veq_dist.R_dist.plot(fig=0)
        plt.yticks([])
        plt.xlabel('Stellar Raidus',labelpad=2)
        plt.xlim(xmin=0)
        plt.sca(ax3)
        self.veq_dist.plot(fig=0,label='V_eq')
        plt.yticks([])
        self.vsini_dist.plot(fig=0,label='VsinI')
        #plt.legend()
        plt.xlabel('Rotational Velocity [km/s]')
        plt.annotate('Veq',xy=(0.05,0.8),color='b',
                     xycoords='axes fraction',
                     fontsize=14)
        plt.annotate('VsinI',xy=(0.05,0.67),color='g',
                     xycoords='axes fraction',
                     fontsize=14)

        plt.subplots_adjust(hspace=.4)

        plt.annotate('$P = {:.2f}\pm{:.2f}$d'.format(self.veq_dist.Prot_dist.mu,
                                                     self.veq_dist.Prot_dist.sig),
                     xy=(0.5,0.3),xycoords='figure fraction',fontsize=24) 


        plt.suptitle(title)

class Cosi_Distribution_FromH5(Cosi_Distribution,dists.Distribution_FromH5):
    def __init__(self,filename,path='',**kwargs):
        dists.Distribution_FromH5.__init__(self,filename,path,
                                     **kwargs)
        self.vsini_dist = dists.Distribution_FromH5(filename,'vsini')
        self.veq_dist = Veq_Distribution_FromH5(filename,'veq')

def diff_Prot_factor(l,alpha=0.23):
    return (1 - alpha*(np.sin(np.deg2rad(l)))**2)

def veq_samples(R_dist,Prot_dist,N=1e4,alpha=0.23,l0=20,sigl=20):
    """Source for diff rot
    """
    ls = stats.norm(l0,sigl).rvs(N)
    Prots = Prot_dist.rvs(N)
    Prots *= diff_Prot_factor(ls,alpha)
    return R_dist.rvs(N)*2*np.pi*RSUN/(Prots*DAY)/1e5 
      
        
class Veq_Distribution(dists.KDE_Distribution):
    def __init__(self,R_dist,Prot_dist,N=1e4,alpha=0.23,l0=20,sigl=20,
                 adaptive=False,bandwidth=0.03,**kwargs):
        self.R_dist = R_dist
        self.Prot_dist = Prot_dist
        self.alpha = alpha
        self.l0 = l0
        self.sigl = sigl
        veqs = veq_samples(R_dist,Prot_dist,N=N,alpha=alpha,l0=l0,sigl=sigl)
        self.samples = veqs

        dists.KDE_Distribution.__init__(self,veqs,adaptive=adaptive,
                                        bandwidth=bandwidth,name='v_eq',
                                        **kwargs)

    def save_hdf(self,filename,path='',**kwargs):
        dists.KDE_Distribution.save_hdf(self,filename,path,**kwargs)
        self.R_dist.save_hdf(filename,'{}/radius'.format(path))
        self.Prot_dist.save_hdf(filename,'{}/Prot'.format(path))

class Veq_Distribution_FromH5(Veq_Distribution,dists.Distribution_FromH5):
    def __init__(self,filename,path='',**kwargs):
        dists.Distribution_FromH5.__init__(self,filename,path,
                                     **kwargs)
        self.R_dist = dists.Distribution_FromH5(filename,'{}/radius'.format(path))
        self.Prot_dist = dists.Distribution_FromH5(filename,'{}/Prot'.format(path))
        if self.minval < 0:
            self.minval = 0

def fveq(z,R,dR,P,dP):
    """z=veq in km/s.  
    """

    R *= 2*pi*RSUN
    dR *= 2*pi*RSUN
    P *= DAY
    dP *= DAY

    exp1 = -P**2/(2*dP**2) - R**2/(2*dR**2)
    exp2 = (dR**2*P + dP**2*R*(z*1e5))**2/(2*dP**2*dR**2*(dR**2 + dP**2*(z*1e5)**2))

    nonexp_term = 2*dP*dR*np.sqrt(dR**2 + dP**2*(z*1e5)**2)

    return 1e5/(4*pi*(dR**2 + dP**2*(z*1e5)**2)**(3/2))*np.exp(exp1 + exp2) *\
        (dR**2 * P*np.sqrt(2*pi) + 
         dP**2 * np.sqrt(2*pi)*R*(z*1e5) + 
         nonexp_term * np.exp(-exp2) +
         np.sqrt(2*pi)*(dR**2*P + dP**2*R*(z*1e5))*erf((dR**2*P + dP**2*R*(z*1e5)) *
                                                       (np.sqrt(2)*dP*dR*
                                                        np.sqrt(dR**2 + dP**2*(z*1e5)**2))))

class Veq_Distribution_Simple(dists.Distribution):
    def __init__(self,R,dR,P,dP):
        self.R = R
        self.dR = dR

        self.P = P
        self.dP = dP

        self.R_dist = dists.Gaussian_Distribution(R,dR)
        self.Prot_dist = dists.Gaussian_Distribution(P,dP)

        self.veq_nominal = (RSUN*2*np.pi*self.R)/(self.P*DAY)/1e5
        self.e_veq_nominal = self.veq_nominal*(np.sqrt((self.dR/self.R)**2 +
                                                       (self.dP/self.P)**2))
        minveq = self.veq_nominal - 5*self.e_veq_nominal
        maxveq = self.veq_nominal + 5*self.e_veq_nominal

        def pdf(x):
            return fveq(x,self.R,self.dR,self.P,self.dP)

        dists.Distribution.__init__(self,pdf,name='Veq',
                                    minval=minveq,maxval=maxveq)

        if self.minval < 0:
            self.minval=0



from __future__ import print_function,division
import logging
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt
from scipy.integrate import quad
import numpy.random as rand
from scipy.misc import logsumexp
from numpy import pi
from scipy.special import erf
from scipy.interpolate import LinearNDInterpolator as interpnd

import pandas as pd

from pkg_resources import resource_filename

from sini_inference_fns import cosi_integrand
#from simpledist import distributions as dists

def fisher(x,k):
    """Fisher distribution 
    """
    return k/(2*np.sinh(k)) * np.exp(k*np.cos(x))*np.sin(x)

def rayleigh(x,sigma):
    """Rayleigh distribution
    """
    return x/sigma**2 * np.exp(-x**2/(2*sigma**2))

def rayleigh_k(x,k):
    """Rayleigh distribution parameterized by Fisher k parameter.
    """
    return k*x * np.exp(-k*x**2/(2))

def sin_fisher(y,k):
    """pdf for y=sin(x) if x is fisher-distributed with parameter k.  Support is [0,1).
    """
    return 1/np.sqrt(1-y**2) * (k/(np.sinh(k)) * y * (np.cosh(k*np.sqrt(1-y**2))))

def cos_uniform(x):
    """ probability distribution of the cosine of a uniform angle, on [0,1]
    """
    return 2./(np.pi*np.sqrt(1-x**2))

    
def cosi_pdf(z,k=1):
    """Equation (11) of Morton & Winn (2014)
    """
    return 2*k/(np.pi*np.sinh(k)) * quad(cosi_integrand,z,1,args=(k,z))[0] #calling cython function

def draw_fisher(n,k=0.1,res=1e4,comp=False):
    psi = np.linspace(0,np.pi,res)
    pdf = fisher(psi,k)
    cdf = pdf.cumsum()
    cdf /= cdf.max()
    r = rand.random(n)
    inds = np.digitize(r,cdf)
    psis = psi[inds]
    if comp:
        return np.pi/2 - psis
    return psi[inds]

RSUN = 6.96e10
DAY = 86400

def generate_veq(R=1.3, dR=0.1, Prot=6, dProt=0.1,nsamples=1e4,plot=False,
                 R_samples=None,Prot_samples=None):
    """ Returns the mean and std equatorial velocity given R,dR,Prot,dProt

    Assumes all distributions are normal.  This will be used mainly for
    testing purposes; I can use MC-generated v_eq distributions when we go for real.
    """
    if R_samples is None:
        R_samples = R*(1 + rand.normal(size=nsamples)*dR)
    else:
        inds = rand.randint(len(R_samples),size=nsamples)
        R_samples = R_samples[inds]

    if Prot_samples is None:
        Prot_samples = Prot*(1 + rand.normal(size=nsamples)*dProt)
    else:
        inds = rand.randint(len(Prot_samples),size=nsamples)
        Prot_samples = Prot_samples[inds]

    veq_samples = 2*np.pi*R_samples*RSUN/(Prot_samples*DAY)/1e5
    
    if plot:
        plt.hist(veq_samples,histtype='step',color='k',bins=50,normed=True)
        d = stats.norm(scale=veq_samples.std(),loc=veq_samples.mean())
        vs = np.linspace(veq_samples.min(),veq_samples.max(),1e4)
        plt.plot(vs,d.pdf(vs),'r')
    
    return veq_samples.mean(),veq_samples.std()

def like_cosi(cosi,vsini_dist,veq_dist,vgrid=None):
    """likelihood of Data (vsini_dist, veq_dist) given cosi
    """
    sini = np.sqrt(1-cosi**2)
    def integrand(v):
        return vsini_dist(v)*veq_dist(v/sini)
    if vgrid is None:
        return quad(integrand,0,np.inf)[0]
    else:
        return np.trapz(integrand(vgrid),vgrid)
    
def cosi_posterior(vsini_dist,veq_dist,vgrid=None,npts=100,vgrid_pts=1000):
    """returns posterior of cosI given dists for vsini and veq (incorporates unc. in vsini)
    """
    if vgrid is None:
        vgrid = np.linspace(vsini_dist.ppf(0.005),vsini_dist.ppf(0.995),vgrid_pts)
    cs = np.linspace(0,1,npts)
    Ls = cs*0
    for i,c in enumerate(cs):
        Ls[i] = like_cosi(c,vsini_dist.pdf,veq_dist.pdf,vgrid=vgrid)
    if np.isnan(Ls[-1]): #hack to prevent nan when cos=1
        Ls[-1] = Ls[-2]
    Ls /= np.trapz(Ls,cs)
    return cs,Ls


def sample_posterior(x,post,nsamples=1):
    """ Returns nsamples from a tabulated posterior (not necessarily normalized)
    """
    cdf = post.cumsum()
    cdf /= cdf.max()
    u = rand.random(size=nsamples)
    inds = np.digitize(u,cdf)
    return x[inds]

def kappa_prior(k):
    return (1+k**2)**(-3/4)

def uniform_samples_from_kappa_prior(kmin,kmax,n,res=1e4):
    ks = np.linspace(kmin,kmax,res)
    pdf = kappa_prior(ks)
    cdf = pdf.cumsum()/pdf.cumsum().max()
    u = np.linspace(0,1,n)
    inds = np.digitize(u,cdf)
    inds = inds.clip(0,len(ks)-1)
    return ks[inds]


def cosi_samples_from_posteriors(vsini_posts,veq_posts,nsamples=100,vsini_upperlim=2):
    samples = np.zeros((len(vsini_posts),nsamples))
    N = len(vsini_posts)
    for i,vsini_post,veq_post in zip(np.arange(len(vsini_posts)),
                                     vsini_posts,veq_posts):
        samples[i,:] = cosi_samples(vsini_dist=vsini_post,
                                    veq_dist=veq_post,nsamples=nsamples)
    return samples


def v_posts_from_dataframe(df,N=1e4,alpha=0.23,l0=20,sigl=20):
    """ names: Prot, e_Prot, R, e_R, vsini, e_vsini
    """
    vsini_posts = []
    veq_posts = []
    if 'ep_R' in df:
        for R,dR_p,dR_m,P,dP,v,dv in zip(df['R'],df['ep_R'],df['em_R'],
                                  df['Prot'],df['e_Prot'],
                                  df['vsini'],df['e_vsini']):
            vsini_posts.append(stats.norm(v,dv))
            if dR_p==dR_m:
                veq_posts.append(Veq_Posterior(R,dR_p,P,dP))
            else:
                R_dist = dists.fit_doublegauss(R,dR_m,dR_p)
                Prot_dist = stats.norm(P,dP)
                veq_posts.append(Veq_Posterior_General(R_dist,Prot_dist,N=N,
                                                       l0=l0,sigl=sigl))
    else:
        for R,dR,P,dP,v,dv in zip(df['R'],df['e_R'],
                                  df['Prot'],df['e_Prot'],
                                  df['vsini'],df['e_vsini']):
            vsini_posts.append(stats.norm(v,dv))
            veq_posts.append(Veq_Posterior(R,dR,P,dP))
        
    return vsini_posts,veq_posts

def all_cosi_samples(vsinis,veqs,dveqs,dvsinis=None,nsamples=100,vsini_upperlim=4):
    if dvsinis is None:
        dvsinis = [None]*len(vsinis)
    if np.size(dvsinis)==1:
        dvsinis = [dvsinis]*len(vsinis)
    samples = np.zeros((len(vsinis),nsamples))
    for i,vsini,dvsini,veq,dveq in zip(range(len(vsinis)),vsinis,dvsinis,veqs,dveqs):
        samples[i,:] = cosi_samples(vsini,veq,dveq,dvsini,nsamples=nsamples,
                                    vsini_upperlim=vsini_upperlim)
    return samples

def build_interpfn(ks=None,nz=100,return_vals=False,filename='cosi_pdf_grid.h5',
                   kmax=300, save=False):
    """
    """
    if ks is None:
        ks = np.concatenate((np.logspace(-2,0,100),np.linspace(1.1,kmax+1,300)))

    nk = len(ks)
    zs = np.linspace(0,0.9999,nz)
    vals = np.zeros(nk*nz)
    allks = vals*0
    allzs = vals*0
    i=0
    for k in ks:
        for z in zs:
            vals[i] = cosi_pdf(z,k)
            allks[i] = k
            allzs[i] = z
            i += 1
    pts = np.array([allzs,allks]).T
    if save:
        df = pd.DataFrame({'z':allzs, 'k':allks, 'val': vals})
        df.to_hdf(filename,'grid')
    if return_vals:
        return pts,vals
    else:
        return interpnd(pts,vals)

def cosi_pdf_fn(filename=resource_filename('obliquity','data/cosi_pdf_grid.h5'),
                recalc=False,**kwargs):
    if recalc:
        return build_interpfn(**kwargs)
    else:
        df = pd.read_hdf(filename,'grid')
        pts = np.array([df['z'],df['k']]).T
        vals = np.array(df['val'])
        return interpnd(pts,vals)

COSI_PDF_FN = cosi_pdf_fn()
#COSI_PDF_FN = build_interpfn()

def lnlike_kappa(k,samples,prior=None):
    if prior is None:
        prior = kappa_prior
    like = COSI_PDF_FN(samples.ravel(),k).reshape(samples.shape)
    return np.log(np.prod(np.sum(like,axis=1)/samples.shape[1])*prior(k))
                  
def kappa_posterior(samples,kmin=0.01,kmax=100,npts=200):
    #ks = uniform_samples_from_kappa_prior(kmin,kmax,npts)
    ks = np.linspace(kmin,kmax,npts)
    lnlikes = ks*0
    for i,k in enumerate(ks):
        lnlikes[i] = lnlike_kappa(k,samples)
    post = np.exp(lnlikes)
    return ks,post/np.trapz(post,ks)


def lnlike_twofisher(f,k1,k2,samples,prior=None):
    if prior is None:
        prior = kappa_prior
    if f < 0 or f > 1:
        return -np.inf
    #if k1 > k2:
    #    return -np.inf
    else:
        term1 = f*COSI_PDF_FN(samples.ravel(),k1)
        term2 = (1-f)*COSI_PDF_FN(samples.ravel(),k2)
        print(term1.shape,term2.shape)
        like = (f*COSI_PDF_FN(samples.ravel(),k1) + 
                (1-f)*COSI_PDF_FN(samples.ravel(),k2)).reshape(samples.shape)
        return np.log(np.prod(np.sum(like,axis=1)/samples.shape[1])*prior(k1)*prior(k2))










from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize
from Cython.Distutils import build_ext
import numpy

def readme():
    with open('README.rst') as f:
        return f.read()

sini_inference_fns = [Extension('sini_inference_fns',['obliquity/sini_inference_fns.pyx'],
                                include_dirs=[numpy.get_include()])]

setup(name = "obliquity",
      version = "0.2",
      description = "Infer the stellar obliquity distribution of transiting planet systems.",
      long_description = readme(),
      author = "Timothy D. Morton",
      author_email = "tim.morton@gmail.com",
      url = "https://github.com/timothydmorton/obliquity",
      packages = find_packages(),
      package_data = {'obliquity': ['data/*']},
      #ext_modules = cythonize(["obliquity/*.pyx"]),
      ext_modules = sini_inference_fns,
      scripts = ['scripts/write_cosi_dist',
                 'scripts/calc_kappa_posterior'],
      cmdclass = {'build_ext': build_ext},
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy'
        ],
      install_requires=['cython','pandas>=0.13','simpledist'],
      zip_safe=False
) 



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.float
ctypedef np.float_t DTYPE_t
cdef DTYPE_t pi = 3.1415926535897932384626433832795

cdef extern from "math.h":
    double sin(double)
    double cos(double)
    double tanh(double)
    double sqrt(double)
    double atan2(double,double)
    double acos(double)
    double abs(double)
    double log(double)
    double ceil(double)
    double cosh(double)

def cosi_integrand(DTYPE_t y, DTYPE_t k, DTYPE_t z):
    """Integrand to Eq. (11) of Morton & Winn (2014)
    """
    return cosh(k*sqrt(1-y*y)) / sqrt(1-y*y) / sqrt(1-(z/y)*(z/y))

